@@ -42,13 +42,18 @@ def read_csv(filename, variable_names=[], validate=True):
     return traj
 
 
-def write_csv(filename, result):
-    """ Save simulation results as a CSV
+def write_csv(filename, result, columns=None):
+    """ Save a simulation result as a CSV file
 
     Parameters:
-        filename  name of the CSV file to write
-        result    structured NumPy array that holds the results
+        filename   name of the CSV file to write
+        result     structured NumPy array that holds the result
+        columns    list of column names to save (None: save all)
     """
+
+    if columns is not None:
+        result = result[['time'] + columns]
+
     header = ','.join(map(lambda s: '"' + s + '"', result.dtype.names))
     np.savetxt(filename, result, delimiter=',', header=header, comments='', fmt='%g')
 
@@ -438,7 +443,7 @@ def fmu_info(filename, causalities=['input', 'output']):
     l.append("")
     l.append("Variables (%s)" % ', '.join(causalities))
     l.append("")
-    l.append('Name                Causality          Start Value  Unit     Description')
+    l.append('Name                Causality              Start Value  Unit     Description')
     for v in md.modelVariables:
         if v.causality not in causalities:
             continue
@@ -456,8 +461,173 @@ def fmu_info(filename, causalities=['input', 'output']):
 
         l.append(v.name + ' ' * (20 - len(v.name)) +
                  v.causality + ' ' * (10 - len(v.causality)) +
-                 ' ' * (20 - len(start)) + start +
+                 ' ' * (24 - len(start)) + start +
                  '  ' + unit + ' ' * (8 - len(unit)) +
                  ' ' + description)
 
     return '\n'.join(l)
+
+
+def visual_c_versions():
+    """ Detect installed Visual C compilers
+
+    Returns: a sorted list of detected Visual C versions e.g. [90, 120, 140]
+    """
+
+    versions = []
+
+    for key in os.environ.keys():
+        if key.upper().startswith('VS') and key.upper().endswith('COMNTOOLS'):
+            versions.append(int(key[len('VS'):-len('COMNTOOLS')]))
+
+    return sorted(versions)
+
+
+def compile_dll(model_description, sources_dir, compiler=None):
+    """ Compile the shared library
+
+    Parameters:
+        sources_dir:    directory that conatains the FMU's source code
+        compiler:       compiler to use (None: use Visual C on Windows, GCC otherwise)
+    """
+
+    from . import platform, sharedLibraryExtension
+
+    if model_description.fmiVersion == '1.0':
+        raise Exception("FMI 1.0 source FMUs are currently not supported")
+
+    if compiler is None:
+        if platform.startswith('win'):
+            compiler = 'vc'
+        else:
+            compiler = 'gcc'
+
+    include_dir = os.path.join(os.path.dirname(__file__), 'c-code')
+    preprocessor_definitions = []
+
+    source_files = []
+
+    if model_description.coSimulation is not None:
+        model_identifier = model_description.coSimulation.modelIdentifier
+        preprocessor_definitions.append('CO_SIMULATION')
+        source_files += model_description.coSimulation.sourceFiles
+
+    if model_description.modelExchange is not None:
+        model_identifier = model_description.coSimulation.modelIdentifier
+
+        for source_file in model_description.modelExchange.sourceFiles:
+            if source_file not in source_files:
+                source_files += source_file
+        preprocessor_definitions.append('MODEL_EXCHANGE')
+
+    if len(source_files) == 0:
+        raise Exception("No source files specified in the model description.")
+
+    source_files += [os.path.join(include_dir, 'fmi2', 'fmi2_wrapper.c')]
+
+    preprocessor_definitions.append('FMI2_FUNCTION_PREFIX=' + model_identifier + '_')
+
+    fmi_include_dir = os.path.join(include_dir, 'fmi1' if model_description.fmiVersion == '1.0' else 'fmi2')
+
+    target = model_identifier + sharedLibraryExtension
+
+    print('Compiling %s...' % target)
+
+    if compiler == 'vc':
+
+        vc_versions = visual_c_versions()
+
+        if len(vc_versions) == 0:
+            raise Exception("No VisualStudio found")
+
+        command = r'call "%%VS%dCOMNTOOLS%%\..\..\VC\vcvarsall.bat"' % vc_versions[-1]
+        if platform == 'win64':
+            command += ' x86_amd64'
+        command += ' && cl /LD /I. /I"%s"' % fmi_include_dir
+        for definition in preprocessor_definitions:
+            command += ' /D' + definition
+        command += ' /Fe' + model_identifier + ' shlwapi.lib ' + ' '.join(source_files)
+
+    elif compiler == 'gcc':
+
+        command = ''
+        if platform.startswith('win'):
+            command += r'set PATH=C:\MinGW\bin;%%PATH%% && '
+        command += 'gcc -c -I. -I%s' % fmi_include_dir
+        if platform in ['linux32', 'linux64']:
+            command += ' -fPIC'
+        for definition in preprocessor_definitions:
+            command += ' -D' + definition
+        command += ' ' + ' '.join(source_files)
+        command += ' && gcc'
+        if platform != 'darwin64':
+            command += ' -static-libgcc'
+        command += ' -shared -o%s *.o' % target
+
+    else:
+        raise Exception("Unsupported compiler: '%s'" % compiler)
+
+    cur_dir = os.getcwd()
+    os.chdir(sources_dir)
+    status = os.system(command)
+    os.chdir(cur_dir)
+
+    print(sources_dir)
+    print(command)
+
+    dll_path = os.path.join(sources_dir, target)
+
+    if status != 0 or not os.path.isfile(dll_path):
+        raise Exception('Failed to compile shared library')
+
+    return str(dll_path)
+
+
+def compile_platform_binary(filename, output_filename=None):
+    """ Compile the binary of an FMU for the current platform and add it to the FMU
+
+    Parameters:
+        filename:         filename of the source code FMU
+        output_filename:  filename of the FMU with the compiled binary (None: overwrite existing FMU)
+    """
+
+    from . import read_model_description, extract, platform
+    import zipfile
+    from shutil import copyfile, rmtree
+
+    unzipdir = extract(filename)
+
+    md = read_model_description(filename)
+
+    binary = compile_dll(model_description=md, sources_dir=os.path.join(unzipdir, 'sources'))
+
+    unzipdir2 = extract(filename)
+
+    platform_dir = os.path.join(unzipdir2, 'binaries', platform)
+
+    # if os.path.exists(platform_dir):
+    #     rmtree(platform_dir)
+
+    if not os.path.exists(platform_dir):
+        os.makedirs(platform_dir)
+
+    copyfile(src=binary, dst=os.path.join(platform_dir, os.path.basename(binary)))
+
+    if output_filename is None:
+        output_filename = filename  # overwrite the existing archive
+
+    # create a new archive from the existing files + compiled binary
+    with zipfile.ZipFile(output_filename, 'w', zipfile.ZIP_DEFLATED) as zf:
+        base_path = os.path.normpath(unzipdir2)
+        for dirpath, dirnames, filenames in os.walk(unzipdir2):
+            for name in sorted(dirnames):
+                path = os.path.normpath(os.path.join(dirpath, name))
+                zf.write(path, os.path.relpath(path, base_path))
+            for name in filenames:
+                path = os.path.normpath(os.path.join(dirpath, name))
+                if os.path.isfile(path):
+                    zf.write(path, os.path.relpath(path, base_path))
+
+    # clean up
+    rmtree(unzipdir)
+    rmtree(unzipdir2)
